// -* coding: utf-8; -*

var Xex = {};

Xex.Error = {
  UnknownError: "unknown-error",
  WrongArgument: "wrong-argument",
  // Load time errors.
  InvalidInteger: "invalid-integer",
  TermTypeInvalid: "term-type-invalid",
  FunctionConflict: "function-conflict",
  VariableTypeConflict: "variable-type-conflict",
  VariableRangeConflict: "variable-range-conflict",
  VariableWrongRange: "variable-wrong-range",
  VariableWrongValue: "variable-wrong-value",

  UnknownFunction: "unknown-function",
  MacroExpansionError: "macro-expansion-error",
  NoVariableName: "no-variable-name",
  NoFunctionName: "no-funcion-name",

  // Run time errors.
  ArithmeticError: "arithmetic-error",
  WrongType: "wrong-type",
  IndexOutOfRange: "index-out-of-range",
  ValueOutOfRange: "value-out-of-range",
  NoLoopToBreak: "no-loop-to-break",
  UncaughtThrow: "uncaught-throw"
};

Xex.Variable = function (domain, name, desc, val, range)
{
  this.domain = domain;
  this.name = name;
  this.desc = desc;
  this.val = val;
  this.range = range;
}

Xex.Variable.prototype.clone = function ()
{
  return new Xex.Variable (this.domain, this.name, this.desc,
			   this.val, this.range);
}
    
Xex.Variable.prototype.Equals = function (obj)
{
  return ((obj instanceof Xex.Variable)
	  && obj.name == this.name);
}

Xex.Variable.prototype.SetValue = function (term)
{
  this.val = term;
  return term;
}

Xex.Function = function (name, with_var, min_args, max_args)
{
  this.name = name;
  this.with_var = with_var;
  this.min_args = min_args;
  this.max_args = max_args;
};  

Xex.Subrountine = function (builtin, name, with_var, min_args, max_args)
{
  this.name = name;
  this.with_var = with_var;
  this.min_args = min_args;
  this.max_args = max_args;
  this.builtin = builtin;
}

Xex.Subrountine.prototype.Call = function (domain, vari, args)
{
  newargs = new Array ();
  for (var i = 0; i < args.length; i++)
    {
      newargs[i] = args[i].Eval (domain);
      if (domain.Thrown ())
	return newargs[i];
    }
  return this.builtin (domain, vari, newargs)
}

Xex.SpecialForm = function (builtin, name, with_var, min_args, max_args)
{
  this.name = name;
  this.with_var = with_var;
  this.min_args = min_args;
  this.max_args = max_args;
  this.builtin = builtin;
}

Xex.SpecialForm.prototype.Call = function (domain, vari, args)
{
  return this.builtin (domain, vari, args)
}

Xex.Lambda = function (name, min_args, max_args, args, body)
{
  this.name = name;
  this.min_args = min_args;
  this.max_args = max_args;
  this.args = args;
  this.body = body;
}

Xex.Lambda.prototype.Call = function (domain, vari, args)
{
  var current = domain.bindings;
  var result = Xex.Zero;
  var limit = max_args >= 0 ? args.length : args.length - 1;
  var i;
  
  try {
    for (i = 0; i < limit; i++)
      {
	result = args[i].Eval (domain);
	if (domain.Thrown ())
	  return result;
	domain.Bind (this.args[i], result);
      }
    if (max_args < 0)
      {
	var list = new Array ();
	for (i = 0; i < args[limit].length; i++)
	  {
	    result = args[limit].Eval (domain);
	    if (domain.Thrown ())
	      return result;
	    list[i] = result;
	  }
	domain.Bind (this.args[limit], list);
      }
    try {
      domain.Catch (Xex.CatchTag.Return);
      for (var term in this.body)
        {
	  result = term.Eval (domain);
	  if (domain.Thrown ())
	    return result;
	}
    } finally {
      domain.Uncatch ();
    }
  } finally {
    domain.UnboundTo (current);
  }
  return result;
}

Xex.Macro = function (name, min_args, max_args, args, body)
{
  this.name = name;
  this.min_args = min_args;
  this.max_args = max_args;
  this.args = args;
  this.body = body;
}

Xex.Macro.prototype.Call = function (domain, vari, args)
{
  var current = domain.bindings;
  var result = Xex.Zero;
  var i;

  try {
    for (i = 0; i < args.length; i++)
      domain.Bind (this.args[i], args[i]);
    try {
      domain.Catch (Xex.CatchTag.Return);
      for (var term in body)
        {
	  result = term.Eval (domain);
	  if (domain.Thrown ())
	    break;
	}
    } finally {
      domain.Uncatch ();
    }
  } finally {
    domain.UnboundTo (current);
  }
  return result;
}

Xex.Bindings = function (vari)
{
  this.vari = vari;
  this.old_value = vari.val;
}

Xex.Bindings.prototype.UnboundTo = function (boundary)
{
  for (var b = this; b != boundary; b = b.next)
    b.vari.val = b.old_value;
  return boundary;
}

Xex.Bind = function (bindings, vari, val)
{
  var b = new Xex.Bindings (vari);
  b.vari.val = val;
  b.next = bindings;
  return b;
}

Xex.CatchTag = {
  Return: 0,
  Break: 1
}

Xex.Domain = function (name, parent, context)
{
  this.name = name;
  this.context = context;
  this.depth = 0;

  if (name != 'basic' && ! parent)
    parent = Xex.BasicDomain
  this.parent = parent;
  this.termtypes = {};
  this.functions = {};
  this.variables = {};
  if (parent)
    {
      var elt;
      for (elt in parent.termtypes)
	this.termtypes[elt] = parent.termtypes[elt];
      for (elt in parent.functions)
	this.functions[elt] = parent.functions[elt];
      for (elt in parent.variables)
	this.variables[elt] = parent.variables[elt];
    }

  this.call_stack = new Array ();
  this.bindings = null;
  this.catch_stack = new Array ();
  this.catch_count = 0;
  this.caught = false;
};

Xex.Domain.prototype = {
  CallStackCount: function () { return this.call_stack.length; },
  CallStackPush: function (term) { this.call_stack.push (term); },
  CallStackPop: function () { this.call_stack.pop (); },
  Bind: function (vari, val)
  {
    this.bindings = Xex.Bind (this.bindings, vari, val);
  },
  UnboundTo: function (boundary)
  {
    if (this.bindings)
      this.bindings = this.bindings.UnboundTo (boundary);
  },
  Catch: function (tag) { this.catch_stack.push (tag); this.catch_count++; },
  Uncatch: function ()
  {
    this.catch_stack.pop ();
    if (this.catch_count > this.catch_stack.length)
      this.catch_count--;
  },
  Thrown: function ()
  {
    if (this.catch_count < this.catch_stack.length)
      {
	this.caught = (this.catch_count == this.catch_stack.length - 1);
	return true;
      }
    this.caught = false;
    return false;
  },
  ThrowReturn: function ()
  {
    for (var i = this.catch_stack.length - 1; i >= 0; i--)
      {
	this.catch_count--;
	if (this.catch_stack[i] == Xex.CatchTag.Return)
	  break;
      }
  },
  ThrowBreak: function ()
  {
    if (this.catch_stack[this.catch_stack.length - 1] != Xex.CatchTag.Break)
      throw new Xex.ErrTerm (Xex.Error.NoLoopToBreak,
			   "No surrounding loop to break");
    this.catch_count--;
  },
  ThrowSymbol: function (tag)
  {
    var i = this.catch_count;
    for (var j = this.catch_stack.length - 1; j >= 0; j--)
      {
	i--;
	if (Xex.CatchTag.Matches (this.catch_stack[i], tag))
          {
	    this.catch_count = i;
	    return;
	  }
      }
    throw new Xex.ErrTerm (Xex.Error.UncaughtThrow,
			 "No corresponding catch: " + tag);
  },
  DefType: function (obj)
  {
    var type = obj.type;
    if (this.termtypes[type])
      throw new Xex.ErrTerm (Xex.Error.TermTypeInvalid,
			   "Already defined: " + type);
    if (this.functions[type])
      throw new Xex.ErrTerm (Xex.Error.TermTypeInvalid,
			   "Already defined as a funciton or a macro: "
			   + type);
    this.termtypes[type] = obj.Parser;
  },
  DefSubr: function (builtin, name, with_var, min_args, max_args)
  {
    this.functions[name] = new Xex.Subrountine (builtin, name, with_var,
						min_args, max_args);
  },
  DefSpecial: function (builtin, name, with_var, min_args, max_args)
  {
    this.functions[name] = new Xex.SpecialForm (builtin, name, with_var,
						min_args, max_args);
  },
  Defun: function (name, min_args, max_args, args, body)
  {
    this.functions[name] =  new Xex.Lambda (name, min_args, max_args,
					    args, body);
  },
  DefunByFunc: function (func) { this.functions[func.name] = func; },
  Defmacro: function (name, min_args, max_args, args, body)
  {
    this.functions[name] = new Xex.Macro (name, min_args, max_args,
					  args, body);
  },
  DefAlias: function (alias, fname)
  {
    var func = this.functions[fname];

    if (! func)
      throw new Xex.ErrTerm (Xex.Error.UnknownFunction, fname);
    if (this.termtypes[alias])
      throw new Xex.ErrTerm (Xex.Error.FunctionConflict,
			   "Already defined as a term type: " + alias);
    if (this.functions[alias])
      throw new Xex.ErrTerm (Xex.Error.FunctionConflict,
			   "Already defined as a function: " + alias);
    this.functions[alias] = func;
  },
  Defvar: function (name, desc, val, range)
  {
    var vari = new Xex.Variable (this, name, desc, val, range);
    this.variables[name] = vari;
    return vari;
  },
  GetFunc: function (name)
  {
    var func = this.functions[name];
    if (! func)
      throw new Xex.ErrTerm (Xex.Error.UnknownFunction,
			     "Unknown function: " + name);
    return func;
  },
  CopyFunc: function (domain, name)
  {
    var func = this.functions[name];
    domain.DefunByFunc (func);
    return true;
  },
  CopyFuncAll: function (domain)
  {
    for (var elt in this.functions)
      domain.DefunByFunc (this.functions[elt]);
  },
  GetVarCreate: function (name)
  {
    var vari = this.variables[name];
    if (! vari)
      vari = this.variables[name] = new Xex.Variable (this, name, null,
						      Xex.Zero, null);
    return vari;
  },
  GetVar: function (name) { return this.variables[name]; },
  SaveValues: function ()
  {
    values = {};
    for (var elt in this.variables)
      {
	//if (! this.variables[elt].val)
	//alert ('unknown value of ' + elt);
	//else
	  values[elt] = this.variables[elt].val.Clone ();
      }
    return values;
  },
  RestoreValues: function (values)
  {
    var name;
    for (name in values)
      {
	var vari = this.variables[name];
	vari.val = values[name];
      }
  }
};

Xex.Term = function (type) { this.type = type; }
Xex.Term.prototype = {
  IsTrue: function () { return true; },
  Eval: function (domain) { return this.Clone (); },
  Clone: function (domain) { return this; },
  Equals: function (obj)
  {
    return (this.type == obj.type
	    && this.val
	    && obj.val == this.val);
  },
  Matches: function (obj) { return this.Equals (obj); },
  toString: function ()
  {
    if (this.val != undefined)
      return '<' + this.type + '>' + this.val + '</' + this.type + '>';
    return '<' + this.type + '/>';
  },
  Intval: function () { throw new Xex.ErrTerm (Xex.Error.WrongType,
					       "Not an integer"); },
  Strval: function () { throw new Xex.ErrTerm (Xex.Error.WrongType,
					       "Not a string"); }
};

Node.prototype.firstElement = function ()
{
  for (var n = this.firstChild; n; n = n.nextSibling)
    if (n.nodeType == 1)
      return n;
  return null;
}

Node.prototype.nextElement = function ()
{
  for (var n = this.nextSibling; n; n = n.nextSibling)
    if (n.nodeType == 1)
      return n;
  return null;
};

(function () {
  function parse_defvar (domain, node)
  {
    var name = node.attributes['vname'].nodeValue;
    if (! name)
      throw new Xex.ErrTerm (Xex.Error.NoVariableName, node, '');
    var vari = domain.variables[name];
    var desc, val, range;
    if (vari)
      {
	desc = vari.description;
	val = vari.val;
	range = vari.range;
      }
    node = node.firstElement ();
    if (node && node.nodeName == 'description')
      {
	desc = node.firstChild.nodeValue;
	node = node.nextElement ();
      }
    if (node)
      {
	val = Xex.Term.Parse (domain, node);
	node = node.nextElement ();
	if (node && node.nodeName == 'possible-values')
	  for (node = node.firstElement (); node; node = node.nextElement ())
	    {
	      var pval;
	      if (node.nodeName == 'range')
		{
		  if (! val.IsInt)
		    throw new Xex.ErrTerm (Xex.Error.TermTypeInvalid,
					   'Range not allowed for ' + name);
		  pval = new Array ();
		  for (var n = node.firstElement (); n; n = n.nextElement ())
		    {
		      var v = Xex.Term.Parse (domain, n);
		      if (! v.IsInt)
			throw new Xex.ErrTerm (Xex.Error.TermTypeInvalid,
					       'Invalid range value: ' + val);
		      pval.push (v);
		    }
		  }
	      else
		{
		  pval = Xex.Term.Parse (domain, node);
		  if (val.type != pval.type)
		    throw new Xex.ErrTerm (Xex.Error.TermTypeInvalid,
					   'Invalid possible value: ' + pval);
		}
	      if (! range)
		range = new Array ();
	      range.push (pval);
	  }
      }
    if (! val)
      val = Xex.Zero;
    domain.Defvar (name, desc, val, range);
    return name;
  }

  function parse_defun_head (domain, node)
  {
    var name = node.attributes['fname'].nodeValue;
    if (! name)
      throw new Xex.ErrTerm (Xex.Error.NoFunctionName, node, '');
    var args = new Array ();
    var nfixed = 0, noptional = 0, nrest = 0;

    node = node.firstElement ();
    if (node && node.nodeName == 'args')
      {
	var n;
	for (n = n.firstElement (); n; n = n.nextElement ())
	  {
	    if (n.nodeName == 'fixed')
	      nfixed++;
	    else if (n.nodeName == 'optional')
	      noptional++;
	    else if (n.nodeName == 'rest')
	      nrest++;
	    else
	      throw new Xex.ErrTerm (Xex.Error.WrongType, n, n.nodeName);
	  }
	if (nrest > 1)
	  throw new Xex.ErrTerm (Xex.Error.WrongType, n, 'Too many <rest>');
	for (n = node.firstElement (); n; n = n.nextElement ())
	  args.push (domain.DefVar (n.attributes['vname'].nodeValue));
      }
    args.min_args = nfixed;
    args.max_args = nrest == 0 ? nfixed + noptional : -1;

    if (node.nodeName == 'defun')
      domain.Defun (name, args, null);
    else
      domain.Defmacro (name, args, null);
    return name;
  }

  function parse_defun_body (domain, node)
  {
    var name = node.attributes['fname'].nodeValue;
    var func = domain.GetFunc (name);
    var body;
    for (node = node.firstElement (); node; node = node.nextElement ())
      if (node.nodeName != 'description' && node.nodeName != 'args')
	break;
    body = Xex.Term.Parse (domain, node, null);
    func.body = body;
  }

  Xex.Term.Parse = function (domain, node, stop)
  {
    if (arguments.length == 2)
      {
	var name = node.nodeName;
	var parser = domain.termtypes[name];

	if (parser)
	  return parser (domain, node);
	if (name == 'defun' || name == 'defmacro')
	  {
	    name = parse_defun_head (domain, node);
	    parse_defun_body (domain, node);
	    return new Xex.StrTerm (name);
	  }
	if (name == 'defvar')
	  {
	    name = parse_defvar (domain, node);
	    return new Xex.StrTerm (name);
	  }
	return new Xex.Funcall.prototype.Parser (domain, node);
      }
    for (var n = node; n && n != stop; n = n.nextElement ())
      if (n.nodeName == 'defun' || n.nodeName == 'defmacro')
	parse_defun_head (domain, n);
    var terms = null;
    for (var n = node; n && n != stop; n = n.nextElement ())
      {
	if (n.nodeName == 'defun' || n.nodeName == 'defmacro')
	  parse_defun_body (domain, n);
	else if (n.nodeName == 'defvar')
	  parse_defvar (domain, n);
	else
	  {
	    if (! terms)
	      terms = new Array ();
	    terms.push (Xex.Term.Parse (domain, n));
	  }
      }
    return terms;
  }
}) ();

Xex.Varref = function (vname)
{
  this.val = vname;
};

(function () {
  var proto = new Xex.Term ('varref');

  proto.Clone = function () { return new Xex.Varref (this.val); }
  proto.Eval = function (domain)
  {
    if (! this.vari || this.vari.domain != domain)
      this.vari = domain.GetVarCreate (this.val);
    return this.vari.val;
  }

  proto.Parser = function (domain, node)
  {
    return new Xex.Varref (node.attributes['vname'].nodeValue);
  }

  Xex.Varref.prototype = proto;
}) ();

var null_args = new Array ();
  
Xex.Funcall = function (func, vari, args)
{
  this.func = func;
  this.vari = vari;
  this.args = args || null_args;
};

(function () {
  var proto = new Xex.Term ('funcall');

  proto.Parser = function (domain, node)
  {
    var fname = node.nodeName;
    var attr;

    if (fname == 'funcall')
      fname = node.attributes['fname'].nodeValue;
    var func = domain.GetFunc (fname);
    var vari;
    attr = node.attributes['vname'];
    vari = attr != undefined ? domain.GetVarCreate (attr.nodeValue) : false;
    var args = Xex.Term.Parse (domain, node.firstElement (), null);
    return new Xex.Funcall (func, vari, args);
  }

  proto.New = function (domain, fname, vname, args)
  {
    var func = domain.GetFunc (fname);
    var vari = vname ? domain.GetVarCreate (vname) : null;
    var funcall = new Xex.Funcall (func, vari, args);
    if (func instanceof Xex.Macro)
      funcall = funcall.Eval (domain);
    return funcall;
  }

  proto.Eval = function (domain)
  {
    return this.func.Call (domain, this.vari, this.args);
  }

  proto.Clone = function ()
  {
    return new Xex.Funcall (this.func, this.vari, this.args);
  }

  proto.Equals = function (obj)
  {
    return (obj.type == 'funcall'
	    && obj.func == this.func
	    && obj.vari.Equals (this.vari)
	    && obj.args.length == this.func.length);
  }

  proto.toString = function ()
  {
    var arglist = ''
    var len = this.args.length;
    if (len == 0)
      return '<' + this.func.name + '/>';
    for (var i = 0; i < len; i++)
      arglist += this.args[i].toString ();
    return '<' + this.func.name + '>' + arglist + '</' + this.func.name + '>';
  }

  Xex.Funcall.prototype = proto;
}) ();

Xex.ErrTerm = function (ename, message, stack)
{
  this.ename = ename;
  this.message = message;
  this.stack = stack;
};

(function () {
  var proto = new Xex.Term ('error');

  proto.IsError = true;

  proto.Parser = function (domain, node)
  {
    return new Xex.ErrTerm (node.attributes['ename'].nodeValue,
			    node.innerText, false);
  }

  proto.CallStack = function () { return stack; }

  proto.SetCallStack = function (value) { statck = value; }

  proto.Clone = function ()
  {
    return new Xex.ErrTerm (ename, message, false);
  }

  proto.Equals = function (obj)
  {
    return (obj.IsError
	    && obj.ename == ename && obj.message == message
	    && (obj.stack ? (stack && stack.length == obj.stack.length)
		: ! stack));
  }

  proto.Matches = function (obj)
  {
    return (obj.IsError && obj.ename == ename);
  }

  proto.toString = function ()
  {
    return '<error ename="' + this.ename + '">' + this.message + '</error>';
  }

  Xex.ErrTerm.prototype = proto;
}) ();

Xex.IntTerm = function (num) { this.val = num; };
(function () {
  var proto = new Xex.Term ('integer');
  proto.IsInt = true;
  proto.Intval = function () { return this.val; };
  proto.IsTrue = function () { return this.val != 0; }
  proto.Clone = function () { return new Xex.IntTerm (this.val); }
  proto.Parser = function (domain, node)
  {
    var str = node.firstChild.nodeValue;

    if (str.charAt (0) == '?' && str.length == 2)
      return new Xex.IntTerm (str.charCodeAt (1));
    return new Xex.IntTerm (parseInt (node.firstChild.nodeValue));
  }
  Xex.IntTerm.prototype = proto;
}) ();

Xex.StrTerm = function (str) { this.val = str; };
(function () {
  var proto = new Xex.Term ('string');
  proto.IsStr = true;
  proto.Strval = function () { return this.val; };
  proto.IsTrue = function () { return this.val.length > 0; }
  proto.Clone = function () { return new Xex.StrTerm (this.val); }
  proto.Parser = function (domain, node)
  {
    return new Xex.StrTerm (node.firstChild.nodeValue);
  }
  Xex.StrTerm.prototype = proto;
}) ();

Xex.SymTerm = function (str) { this.val = str; };
(function () {
  var proto = new Xex.Term ('symbol');
  proto.IsSymbol = true;
  proto.IsTrue = function () { return this.val != 'nil'; }
  proto.Clone = function () { return new Xex.SymTerm (this.val); }
  proto.Parser = function (domain, node)
  {
    return new Xex.SymTerm (node.firstChild.nodeValue);
  }
  Xex.SymTerm.prototype = proto;
}) ();

Xex.LstTerm = function (list) { this.val = list; };
(function () {
  var proto = new Xex.Term ('list');
  proto.IsList = true;
  proto.IsTrue = function () { return this.val.length > 0; }
  proto.Clone = function () { return new Xex.LstTerm (this.val.slice (0)); }

  proto.Equals = function (obj)
  {
    if (obj.type != 'list' || obj.val.length != this.val.length)
      return false;
    var i, len = this.val.length;
    for (i = 0; i < len; i++)
      if (! this.val[i].Equals (obj.val[i]))
	return false;
    return true;
  }

  proto.Parser = function (domain, node)
  {
    var list = Xex.Term.Parse (domain, node.firstElement (), null);
    return new Xex.LstTerm (list);
  }

  proto.toString = function ()
  {
    var len = this.val.length;

    if (len == 0)
      return '<list/>';
    var str = '<list>';
    for (var i = 0; i < len; i++)
      str += this.val[i].toString ();
    return str + '</list>';
  }
  Xex.LstTerm.prototype = proto;
}) ();

(function () {
  var basic = new Xex.Domain ('basic', null, null);

  function Fset (domain, vari, args)
  {
    return vari.SetValue (args[0]);
  }

  function maybe_set_intvar (vari, n)
  {
    var term = new Xex.IntTerm (n);
    if (vari != null)
      vari.SetValue (term);
    return term;
  }

  function Fset (domain, vari, args)
  {
    if (! vari)
      throw new Xex.ErrTerm (Xex.Error.NoVariableName,
			     'No variable name to set');
    vari.SetValue (args[0]);
    return args[0];
  }

  function maybe_set_intvar (vari, n)
  {
    var term = new IntTerm (n);
    if (vari)
      vari.SetValue (term);
    return term;
  }

  function Fadd (domain, vari, args)
  {
    var n = vari ? vari.val.Intval () : 0;
    var len = args.length;

    for (var i = 0; i < len; i++)
      n += args[i].Intval ();
    return maybe_set_intvar (vari, n);
  }

  function Fmul (domain, vari, args)
  {
    var n = vari ? vari.val.Intval () : 1;
    for (var i = 0; i < args.length; i++)
      n *= arg.Intval ();
    return maybe_set_intvar (vari, n);
  }

  function Fsub (domain, vari, args)
  {
    var n, i;

    if (! vari)
      {
	n = args[0].Intval ();
	i = 1;
      }
    else
      {
	n = vari.val.Intval ();
	i = 0;
      }
    while (i < args.length)
      n -= args[i++].Intval ();
    return maybe_set_intvar (vari, n);
  }

  function Fdiv (domain, vari, args)
  {
    var n, i;

    if (! vari == null)
      {
	n = args[0].Intval ();
	i = 1;
      }
    else
      {
	n = vari.val.Intval ();
	i = 0;
      }
    while (i < args.length)
      n /= args[i++].Intval ();
    return maybe_set_intvar (vari, n);
  }

  function Fmod (domain, vari, args)
  {
    return maybe_set_intvar (vari, args[0].Intval () % args[1].Intval ());
  }

  function Flogior (domain, vari, args)
  {
    var n = vari == null ? 0 : vari.val;
    for (var i = 0; i < args.length; i++)
      n |= args[i].val;
    return maybe_set_intvar (vari, n);
  }

  function Fand (domain, vari, args)
  {
    var len = args.length;
    for (var i = 0; i < len; i++)
    {
      var result = args[i].Eval (domain);
      if (domain.Thrown ())
	return result;
      if (! result.IsTrue ())
	return Xex.Zero;
    }
    return Xex.One;
  }

  function For (domain, vari, args)
  {
    var len = args.length;
    for (var i = 0; i < len; i++)
    {
      var result = args[i].Eval (domain);
      if (domain.Thrown ())
	return result;
      if (result.IsTrue ())
	return Xex.One;
    }
    return Xex.Zero;
  }

  function Feq (domain, vari, args)
  {
    for (var i = 1; i < args.length; i++)
      if (! args[i - 1].Equals (args[i]))
	return Xex.Zero;
    return Xex.One;
  }

  function Flt (domain, vari, args)
  {
    var n = args[0].Intval;

    for (var i = 1; i < args.length; i++)
      {
	var n1 = args[i].Intval;
	if (n >= n1)
	  return Xex.Zero;
	n = n1;
      }
    return Xex.One;
  }

  function Fle (domain, vari, args)
  {
    var n = args[0].Intval;
    for (var i = 1; i < args.length; i++)
      {
	var n1 = args[i].Intval;
	if (n > n1)
	  return Xex.Zero;
	n = n1;
      }
    return Xex.One;
  }

  function Fgt (domain, vari, args)
  {
    var n = args[0].Intval;
    for (var i = 1; i < args.length; i++)
      {
	var n1 = args[i].Intval;
	if (n <= n1)
	  return Xex.Zero;
	n = n1;
      }
    return Xex.One;
  }

  function Fge (domain, vari, args)
  {
    var n = args[0].Intval;
    for (var i = 1; i < args.Length; i++)
      {
	var n1 = args[i].Intval;
	if (n < n1)
	  return Xex.Zero;
	n = n1;
      }
    return Xex.One;
  }

  function Fprogn (domain, vari, args)
  {
    var result = Xex.One;
    var len = args.length;

    for (var i = 0; i < len; i++)
      {
	result = args[i].Eval (domain);
	if (domain.Thrown ())
	  return result;
      }
    return result;
  }

  function Fif (domain, vari, args)
  {
    var result = args[0].Eval (domain);

    if (domain.Thrown ())
      return result;
    if (result.IsTrue ())
      return args[1].Eval (domain);
    if (args.length == 2)
      return Zero;
    return args[2].Eval (domain);
  }

  function Fcond (domain, vari, args)
  {
    for (var i = 0; i < args.length; i++)
      {
	var list = args[i].val;
	var result = list.val[0].Eval (doamin);
	if (result.isTrue ())
	  {
	    for (var j = 1; j < list.val.length; j++)
	      {
		domain.depth++;
		result = list.val[j].Eval (domain);
		domain.depth--;
		if (domain.Thrown ())
		  return result;
		}
	    return result;
	  }
      }
    return Xex.Zero;
  }

  function eval_terms (domain, terms, idx)
  {
    var result = Xex.Zero;
    domain.caught = false;
    for (var i = idx; i < terms.length; i++)
      {
	result = terms[i].Eval (domain);
	if (domain.Thrown ())
	  return result;
      }
    return result;
  }

  function Fcatch (domain, vari, args)
  {
    var caught = false;
    var result;

    if (args[0].IsError)
      {
	try {
	  result = eval_terms (domain, args, 1);
	} catch (e) {
	  if (e instanceof Xex.ErrTerm)
	    {
	      if (! args[0].Matches (e))
		throw e;
	      if (vari)
		vari.SetValue (e);
	      return Xex.One;
	    }
	}
      }
    else if (args[0].IsSymbol)
      {
	try {
	  domain.Catch (args[0].val);
	  result = eval_terms (domain, args, 1);
	  if (domain.caught)
	    {
	      if (vari != null)
		vari.SetValue (result);
	      return Xex.One;
	    }
	  return Xex.Zero;
	} finally {
	  domain.Uncatch ();
	}
      }
    throw new Xex.ErrTerm (Xex.Error.WrongArgument,
			   "Not a symbol nor an error: " + args[0]);
  }

  function Fthrow (domain, vari, args)
  {
    if (args[0].IsSymbl)
      {
	domain.ThrowSymbol (args[0]);
	return (args[args.length - 1]);
      }
    if (args[0].IsError)
      {
	throw args[0];
      }
    throw new Xex.ErrTerm (Xex.Error.WrongArgument,
			   "Not a symbol nor an error:" + args[0]);
  }

  Xex.BasicDomain = basic;

  basic.DefSubr (Fset, "set", true, 1, 1);
  if (basic.functions['='])
    alert (basic.functions['=']);
  basic.DefAlias ("=", "set");
  //basic.DefSubr (Fnot, "not", false, 1, 1);
  //basic.DefAlias ("!", "not");
  basic.DefSubr (Fadd, "add", true, 1, -1);
  basic.DefSubr (Fmul, "mul", true, 1, -1);
  basic.DefAlias ("*", "mul");
  basic.DefSubr (Fsub, "sub", true, 1, -1);
  basic.DefAlias ("-", "sub");
  basic.DefSubr (Fdiv, "div", true, 1, -1);
  basic.DefAlias ("/", "div");
  basic.DefSubr (Fmod, "mod", true, 1, 2);
  basic.DefAlias ("%", "mod");
  basic.DefSubr (Flogior, "logior", true, 1, -1);
  basic.DefAlias ('|', "logior");
  //basic.DefSubr (Flogand, "logand", true, 1, -1);
  //basic.DefAlias ("&", "logand");
  //basic.DefSubr (Flsh, "lsh", true, 1, 2);
  //basic.DefAlias ("<<", "lsh");
  //basic.DefSubr (Frsh, "rsh", true, 1, 2);
  //basic.DefAlias (">>", "rsh");
  basic.DefSubr (Feq, "eq", false, 2, -1);
  basic.DefAlias ("==", "eq");
  //basic.DefSubr (Fnoteq, "noteq", false, 2, 2);
  //basic.DefAlias ("!=", "noteq");
  basic.DefSubr (Flt, "lt", false, 2, -1);
  basic.DefAlias ("<", "lt");
  basic.DefSubr (Fle, "le", false, 2, -1);
  basic.DefAlias ("<=", "le");
  basic.DefSubr (Fgt, "gt", false, 2, -1);
  basic.DefAlias (">", "gt");
  basic.DefSubr (Fge, "ge", false, 2, -1);
  basic.DefAlias (">=", "ge");
  basic.DefSubr (Fthrow, "throw", false, 1, 2);

  //basic.DefSubr (Fappend, "append", true, 0, -1);
  //basic.DefSubr (Fconcat, "concat", true, 0, -1);
  //basic.DefSubr (Fnth, "nth", false, 2, 2);
  //basic.DefSubr (Fcopy, "copy", false, 1, 1);
  //basic.DefSubr (Fins, "ins", true, 2, 2);
  //basic.DefSubr (Fdel, "del", true, 2, 2);
  //basic.DefSubr (Feval, "eval", false, 1, 1);
  //basic.DefSubr (Fbreak, "break", false, 0, 1);
  //basic.DefSubr (Freturn, "return", false, 0, 1);
  //basic.DefSubr (Fthrow, "throw", false, 1, 2);

  basic.DefSpecial (Fand, "and", false, 1, -1);
  basic.DefAlias ("&&", "and");
  basic.DefSpecial (For, "or", false, 1, -1);
  basic.DefAlias ("||", "or");
  basic.DefSpecial (Fprogn, "progn", false, 1, -1);
  basic.DefAlias ("expr", "progn");
  basic.DefSpecial (Fif, "if", false, 2, 3);
  //basic.DefSpecial (Fwhen, "when", false, 1, -1);
  //basic.DefSpecial (Floop, "loop", false, 1, -1);
  //basic.DefSpecial (Fwhile, "while", false, 1, -1);
  basic.DefSpecial (Fcond, "cond", false, 1, -1);
  //basic.DefSpecial (Fforeach, "foreach", true, 2, -1);
  //basic.DefSpecial (Fquote, "quote", false, 1, 1);
  //basic.DefSpecial (Ftype, "type", false, 1, 1);
  basic.DefSpecial (Fcatch, "catch", true, 2, -1);

  basic.DefType (Xex.Funcall.prototype);
  basic.DefType (Xex.Varref.prototype);
  basic.DefType (Xex.ErrTerm.prototype);
  basic.DefType (Xex.IntTerm.prototype);
  basic.DefType (Xex.StrTerm.prototype);
  basic.DefType (Xex.SymTerm.prototype);
  basic.DefType (Xex.LstTerm.prototype);

}) ();

Xex.Zero = new Xex.IntTerm (0);
Xex.One = new Xex.IntTerm (1);
Xex.nil = new Xex.SymTerm ('nil');

Xex.Load = function (server, file)
{
  var obj = new XMLHttpRequest ();
  var url = server ? server + '/' + file : file;
  obj.open ('GET', url, false);
  obj.overrideMimeType ('text/xml');
  obj.send ('');
  return obj.responseXML.firstChild;
}

var MIM = {
  // URL of the input method server.
  server: "http://www.m17n.org/common/mim-js",
  // Boolean flag to tell if MIM is active or not.
  enabled: true,
  // Boolean flag to tell if MIM is running in debug mode or not.
  debug: false,
  // List of main input methods.
  imlist: {},
  // List of extra input methods;
  imextra: {},
  // Global input method data
  im_global: null,
  // Currently selected input method.
  current: false,

  // enum
  LoadStatus: { NotLoaded:0, Loading:1, Loaded:2, Error:-1 },
  ChangedStatus: {
    None:	0x00,
    StateTitle:	0x01,
    PreeditText:0x02,
    CursorPos:	0x04,
    CandidateList:0x08,
    CandidateIndex:0x10,
    CandidateShow:0x20,
    Preedit:	0x06,		// PreeditText | CursorPos
    Candidate:	0x38 // CandidateList | CandidateIndex | CandidateShow
  },
  KeyModifier: {
    SL:	0x00400000,
    SR:	0x00800000,
    S:	0x00C00000,
    CL:	0x01000000,
    CR:	0x02000000,
    C:	0x03000000,
    AL:	0x04000000,
    AR:	0x08000000,
    A:	0x0C000000,
    ML:	0x04000000,
    MR:	0x08000000,
    M:	0x0C000000,
    G:	0x10000000,
    s:	0x20000000,
    H:	0x40000000,
    High:	0x70000000,
    All:	0x7FC00000
  },
  Error: {
    ParseError: "parse-error"
  }
};
  
(function () {
  var keysyms = new Array ();
  keysyms["bs"]	= "backspace";
  keysyms["lf"]	= "linefeed";
  keysyms["cr"]	= keysyms["enter"] = "return";
  keysyms["esc"] = "escape";
  keysyms["spc"] = "space";
  keysyms["del"] = "delete";

  function decode_keysym (str) {
    var parts = str.split ("-");
    var len = parts.length, i;
    var has_modifier = len > 1;

    for (i = 0; i < len - 1; i++)
      if (! MIM.KeyModifier.hasOwnProperty (parts[i]))
	return false;
    var key = parts[len - 1];
    if (key.length > 1)
      {
	key = keysyms[key.toLowerCase ()];
	if (key)
	  {
	    if (len > 1)
	      {
		str = parts[0];
		for (i = 1; i < len - 1; i++)
		  str += '-' + parts[i];
		str += '-' + key;
	      }
	    else
	      str = key;
	  }
      }
    if (has_modifier)
      {
	parts = new Array ();
	parts.push (str);
	return parts;
      }
    return str;
  }

  MIM.Key = function (val)
  {
    this.key;
    this.has_modifier = false;
    if (typeof val == 'string' || val instanceof String)
      {
	this.key = decode_keysym (val);
	if (! this.key)
	  throw new Xex.ErrTerm (MIM.Error.ParseError, "Invalid key: " + val);
	if (this.key instanceof Array)
	  {
	    this.key = this.key[0];
	    this.has_modifier = true;
	  }
      }
    else if (typeof val == 'number' || val instanceof Number)
      this.key = String.fromCharCode (val);
    else
      throw new Xex.ErrTerm (MIM.Error.ParseError, "Invalid key: " + val);
  }

  MIM.Key.prototype.toString = function () { return this.key; };
}) ();

(function () {
  MIM.KeySeq = function (seq)
  {
    this.val = new Array ();
    this.has_modifier = false;

    if (seq)
      {
	if (seq.IsList)
	  {
	    var len = seq.val.length;
	    for (var i = 0; i < len; i++)
	      {
		var v = seq.val[i];
		if (v.type != 'string' && v.type != 'integer'
		    && v.type != 'symbol')
		  throw new Xex.ErrTerm (MIM.Error.ParseError,
					 "Invalid key: " + v);
		var key = new MIM.Key (v.val);
		this.val.push (key);
		if (key.has_modifier)
		  this.has_modifier = true;
	      }
	  }
	else if (seq.IsStr)
	  {
	    var len = seq.val.length;
	    for (var i = 0; i < len; i++)
	      this.val.push (new MIM.Key (seq.val.charCodeAt (i)));
	  }
	else
	  throw new Xex.ErrTerm (MIM.Error.ParseError, "Invalid key: " + seq);
      }
  }

  var proto = new Xex.Term ('keyseq');
  proto.Clone = function () { return this; }
  proto.Parser = function (domain, node)
  {
    var seq = new Array ();
    for (node = node.firstChild; node; node = node.nextSibling)
      if (node.nodeType == 1)
	{
	  var term = Xex.Term.Parse (domain, node);
	  return new MIM.KeySeq (term);
	}
    throw new Xex.ErrTerm (MIM.Error.ParseError, "Invalid keyseq");
  }
  proto.toString = function ()
  {
    var len = this.val.length;
    if (len == 0)
      return '<keyseq/>';
    var first = true;
    var str = '<keyseq>';
    for (var i = 0; i < len; i++)
      {
	if (first)
	  first = false;
	else if (this.has_modifier)
	  str += ' ';
	str += this.val[i].toString ();
      }
    return str + '</keyseq>';
  }

  MIM.KeySeq.prototype = proto;
}) ();

(function () {
  MIM.Marker = function () { }
  MIM.Marker.prototype = new Xex.Term ('marker');
  MIM.Marker.prototype.CharAt = function (ic)
  {
    var p = this.Position (ic);
    if (p < 0)
      return ic.GetSurroundingChar (p);
    else if (p >= ic.preedit.length)
      return ic.GetSurroundingChar (p - ic.preedit.length);
    return ic.preedit.charCodeAt (p);
  }

  MIM.NamedMarker = function (name) { this.val = name; }
  MIM.NamedMarker.prototype = new MIM.Marker ();
  MIM.NamedMarker.prototype.Position = function (ic)
  {
    var p = ic.marker_positions[this.val];
    return (p == undefined ? 0 : p);
  }
  MIM.NamedMarker.prototype.Mark = function (ic)
  {
    ic.marker_positions[this.val] = ic.cursor_pos;
  }

  MIM.PredefinedMarker = function (name) { this.val = name; }
  MIM.PredefinedMarker.prototype = new MIM.Marker ();
  MIM.PredefinedMarker.prototype.Position = function (ic)
  {
    if (typeof this.pos == 'number')
      return this.pos;
    return this.pos (ic);
  }

  var predefined = { }

  function def_predefined (name, position)
  {
    predefined[name] = new MIM.PredefinedMarker (name);
    predefined[name].pos = position;
  }

  def_predefined ('@<', 0);
  def_predefined ('@>', function (ic) { return ic.preedit.length; });
  def_predefined ('@-', function (ic) { return ic.cursor_pos - 1; });
  def_predefined ('@+', function (ic) { return ic.cursor_pos + 1; });
  def_predefined ('@[', function (ic) {
    if (ic.cursor_pos > 0)
      {
	var pos = ic.cursor_pos;
	return ic.preedit.FindProp ('candidates', pos - 1).from;
      }
    return 0;
  });
  def_predefined ('@]', function (ic) {
    if (ic.cursor_pos < ic.preedit.length - 1)
      {
	var pos = ic.cursor_pos;
	return ic.preedit.FindProp ('candidates', pos).to;
      }
    return ic.preedit.length;
  });
  for (var i = 0; i < 10; i++)
    def_predefined ("@" + i, i);
  predefined['@first'] = predefined['@<'];
  predefined['@last'] = predefined['@>'];
  predefined['@previous'] = predefined['@-'];
  predefined['@next'] = predefined['@+'];
  predefined['@previous-candidate-change'] = predefined['@['];
  predefined['@next-candidate-change'] = predefined['@]'];

  MIM.SurroundMarker = function (name)
  {
    this.val = name;
    this.distance = parseInt (name.slice (2));
    if (isNaN (this.distance))
      throw new Xex.ErrTerm (MIM.Error.ParseError, "Invalid marker: " + name);
  }
  MIM.SurroundMarker.prototype = new MIM.Marker ();
  MIM.SurroundMarker.prototype.Position = function (ic)
  {
    return ic.cursor_pos + this.distance;
  }

  MIM.Marker.prototype.Parser = function (domain, node)
  {
    var name = node.firstChild.nodeValue;
    if (name.charAt (0) == '@')
      {
	var n = predefined[name];
	if (n)
	  return n;
	if (name.charAt (1) == '-')
	  return new MIM.SurroundMarker (name);
	throw new Xex.ErrTerm (MIM.Error.ParseError,
			       "Invalid marker: " + name);
      }
    return new MIM.NamedMarker (name);
  }
}) ();

MIM.Selector = function (name)
{
  this.val = name;
}
MIM.Selector.prototype = new Xex.Term ('selector');

(function () {
  var selectors = {};
  selectors["@<"] = selectors["@first"] = new MIM.Selector ('@<');
  selectors["@="] = selectors["@current"] = new MIM.Selector ('@=');
  selectors["@>"] = selectors["@last"] = new MIM.Selector ('@>');
  selectors["@-"] = selectors["@previous"] = new MIM.Selector ('@-');
  selectors["@+"] = selectors["@next"] = new MIM.Selector ('@+');
  selectors["@["] = selectors["@previous-candidate-change"]
    = new MIM.Selector ('@[');
  selectors["@]"] = selectors["@next-candidate-change"]
    = new MIM.Selector ('@]');

  MIM.Selector.prototype.Parser = function (domain, node)
  {
    var name = node.firstChild.nodeValue;
    var s = selectors[name];
    if (! s)
      throw new Xex.ErrTerm (MIM.Error.ParseError,
			     "Invalid selector: " + name);
    return s;
  }
}) ();

MIM.Rule = function (keyseq, actions)
{
  this.keyseq = keyseq;
  this.actions = actions;
}
MIM.Rule.prototype = new Xex.Term ('rule');
MIM.Rule.prototype.Parser = function (domain, node)
{
  var n;
  for (n = node.firstChild; n && n.nodeType != 1; n = n.nextSibling);
  if (! n)
    throw new Xex.ErrTerm (MIM.Error.ParseError, "invalid rule:" + node);
  var keyseq = Xex.Term.Parse (domain, n);
  if (keyseq.type != 'keyseq')
    throw new Xex.ErrTerm (MIM.Error.ParseError, "invalid rule:" + node);
  var actions = Xex.Term.Parse (domain, n.nextElement (), null);
  return new MIM.Rule (keyseq, actions);
}
MIM.Rule.prototype.toString = function ()
{
  return '<rule/>';
}

MIM.Map = function (name)
{
  this.name = name;
  this.rules = new Array ();
};

(function () {
  var proto = new Xex.Term ('map');

  proto.Parser = function (domain, node)
  {
    var name = node.attributes['mname'].nodeValue;
    if (! name)
      throw new Xex.ErrTerm (MIM.Error.ParseError, "invalid map");
    var map = new MIM.Map (name);
    for (var n = node.firstChild; n; n = n.nextSibling)
      if (n.nodeType == 1)
	map.rules.push (Xex.Term.Parse (domain, n));
    return map;
  }

  proto.toString = function ()
  {
    var str = '<map mname="' + this.name + '">';
    var len = this.rules.length;
    for (i = 0; i < len; i++)
      str += this.rules[i];
    return str + '</map>';
  }

  MIM.Map.prototype = proto;
}) ();

Xex.CatchTag._mimtag = new Xex.SymTerm ('@mimtag');

MIM.Action = function (domain, terms)
{
  var args = new Array ();
  args.push (Xex.CatchTag_.mimtag);
  for (var i = 0; i < terms.length; i++)
    args.push (terms[i]);
  this.action = Xex.Funcall.prototype.New (domain, 'catch', null, args);
}

MIM.Action.prototype.Run = function (domain)
{
  var result = this.action.Eval (domain);
  if (result.type == 'error')
    {
      domain.context.Error = result.toString ();
      return false;
    }
  return (result != Xex.CatchTag._mimtag);
}

MIM.Keymap = function ()
{
  this.name = 'TOP';
  this.submaps = null;
  this.actions = null;
};

(function () {
  var proto = {};

  function add_rule (keymap, rule)
  {
    var keyseq = rule.keyseq;
    var len = keyseq.val.length;
    var name = '';

    for (var i = 0; i < len; i++)
      {
	var key = keyseq.val[i];
	var sub = false;

	name += key.key;
	if (! keymap.submaps)
	  keymap.submaps = {};
	else
	  sub = keymap.submaps[key.key];
	if (! sub)
	  keymap.submaps[key.key] = sub = new MIM.Keymap ();
	keymap = sub;
	keymap.name = name;
      }
    keymap.actions = rule.actions;
  }

  proto.Add = function (map)
  {
    var rules = map.rules;
    var len = rules.length;

    for (var i = 0; i < len; i++)
      add_rule (this, rules[i]);
  }
  proto.Lookup = function (keys, index)
  {
    var sub;

    if (index < keys.val.length && this.submaps
	&& (sub = this.submaps[keys.val[index].key]))
      {
	index++;
	return sub.Lookup (keys, index);
      }
    return { map: this, index: index };
  }

  MIM.Keymap.prototype = proto;
}) ();

MIM.State = function (name)
{
  this.name = name;
  this.keymap = new MIM.Keymap ();
};

(function () {
  var proto = new Xex.Term ('state');

  proto.Parser = function (domain, node)
  {
    var map_list = domain.map_list;
    var name = node.attributes['sname'].nodeValue;
    if (! name)
      throw new Xex.ErrTerm (MIM.Error.ParseError, "invalid map");
    var state = new MIM.State (name);
    for (node = node.firstElement (); node; node = node.nextElement ())
      {
	if (node.nodeName == 'title')
	  state.title = node.firstChild.nodeValue;
	else
	  {
	    var n = node.firstElement ();
	    if (node.nodeName == 'branch')
	      {
		state.keymap.Add (map_list[node.attributes['mname'].nodeValue]);
		state.keymap.actions = Xex.Term.Parse (domain, n, null);
	      }
	    else if (node.nodeName == 'state-hook')
	      state.enter_actions = Xex.Term.Parse (domain, n, null);
	    else if (node.nodeName == 'catch-all-branch')
	      state.fallback_actions = Xex.Term.Parse (domain, n, null);
	  }
      }
    return state;
  }

  proto.toString = function ()
  {
    return '<state sname="' + this.name + '">' + this.keymap + '</state>';
  }

  MIM.State.prototype = proto;
}) ();

MIM.im_domain = new Xex.Domain ('input-method', null, null);
MIM.im_domain.DefType (MIM.KeySeq.prototype);
MIM.im_domain.DefType (MIM.Marker.prototype);
MIM.im_domain.DefType (MIM.Selector.prototype);
MIM.im_domain.DefType (MIM.Rule.prototype);
MIM.im_domain.DefType (MIM.Map.prototype);
MIM.im_domain.DefType (MIM.State.prototype);

(function () {
  var im_domain = MIM.im_domain;

  function Finsert (domain, vari, args)
  {
    var text;
    if (args[0].type == 'integer')
      text = String.fromCharCode (args[0].val);
    else
      text = args[0].val;
    domain.context.insert (text, null);
  }

  function Finsert_candidates (domain, vari, args)
  {
    var ic = domain.context;
    var gsize = domain.variables['candidates_group_size'];
    var candidates = new MIM.Candidates (args, gsize ? gsize.Intval : 0);
    ic.insert (candidates.Current (), candidates);
    return args[0];
  }

  function Fdelete (domain, vari, args)
  {
    var ic = domain.context;
    var pos = args[0].IsInt ? args[0].Intval : args[0].Position (ic);
    ic.del (pos);
    return new Xex.Term (ic.del (pos));
  }

  function Fselect (domain, vari, args)
  {
    var ic = domain.context;
    var can = ic.candidates;

    if (can)
      {
	var candidate = can.Current ();

	ic.del (ic.cursor_pos - candidate.length);
	candidate = can.Select (args[0]);
	ic.insert (candidate, can);
      }
    return args[0];
  }

  function Fchar_at (domain, vari, args)
  {
    return new Xex.Term (args[0].CharAt (domain.context));
  }

  function Fmove (domain, vari, args)
  {
    var ic = domain.context;
    var pos = args[0].IsInt ? args[0].val : args[0].Position (ic);
    ic.move (pos);
    return args[0];
  }

  function Fmark (domain, vari, args)
  {
    args[0].Mark (domain.context);
    return args[0];
  }

  function Fpushback (domain, vari, args)
  {
    var arg = (args[0].IsInt ? args[0].Intval
	       : args[0].IsStr ? new KeySeq (args[0])
	       : args[0]);
    domain.context.pushback (arg)
    return args[0];
  }

  function Fundo  (domain, vari, args)
  {
    var ic = domain.context;
    var n = args.length == 0 ? -2 : args[0].val;
    if (n < 0)
      ic.keys.val.splice (ic.keys.length + n, -n);
    else
      ic.keys.val.splice (n, ic.keys.length);
    ic.reset ();
    return Xex.nil;
  }

  function Fcommit (domain, vari, args)
  {
    domain.context.commit ();
    return Xex.nil;
  }

  function Funhandle (domain, vari, args)
    {
      domain.context.commit ();
      return Xex.Fthrow (domain, vari, Xex.CatchTag._mimtag);
    }

  function Fshift (domain, vari, args)
  {
    var ic = domain.context;
    var state_name = args[0].val;
    var state = ic.im.state_list[state_name];
    if (! state)
      throw ("Unknown state: " + state_name);
      ic.shift (state);
    return args[0];
  }

  function Fsurrounding_flag (domain, vari, args)
  {
    return new Xex.IntTerm (-1);
  }

  im_domain.DefSubr (Finsert, "insert", false, 1, 1);
  im_domain.DefSubr (Finsert_candidates, "insert-candidates", false, 1, 1);
  im_domain.DefSubr (Fdelete, "delete", false, 1, 1);
  im_domain.DefSubr (Fselect, "select", false, 1, 1);
  //im_domain.DefSubr (Fshow, "show-candidates", false, 0, 0);
  //im_domain.DefSubr (Fhide, "hide-candidates", false, 0, 0);
  im_domain.DefSubr (Fmove, "move", false, 1, 1);
  im_domain.DefSubr (Fmark, "mark", false, 1, 1);
  im_domain.DefSubr (Fpushback, "pushback", false, 1, 1);
  //im_domain.DefSubr (Fpop, "pop", false, 0, 0);
  im_domain.DefSubr (Fundo, "undo", false, 0, 1);
  im_domain.DefSubr (Fcommit, "commit", false, 0, 0);
  im_domain.DefSubr (Funhandle, "unhandle", false, 0, 0);
  im_domain.DefSubr (Fshift, "shift", false, 1, 1);
  //im_domain.DefSubr (Fshiftback, "shiftback", false, 0, 0);
  im_domain.DefSubr (Fchar_at, "char-at", false, 1, 1);
  //im_domain.DefSubr (Fkey_count, "key-count", false, 0, 0);
  im_domain.DefSubr (Fsurrounding_flag, "surrounding-text-flag", false, 0, 0);
}) ();


(function () {
  function get_global_var (vname)
  {
    if (MIM.im_global.load_status == MIM.LoadStatus.NotLoaded)
      MIM.im_global.Load ()
    return MIM.im_global.domain.variables[vname];
  }

  function include (node)
  {
    node = node.firstElement ();
    if (node.nodeName != 'tags')
      return null;
    
    var lang = null, name = null, extra = null;
    for (node = node.firstElement (); node; node = node.nextElement ())
      {
	if (node.nodeName == 'language')
	  lang = node.firstChild.nodeValue;
	else if (node.nodeName == 'name')
	  name = node.firstChild.nodeValue;
	else if (node.nodeName == 'extra-id')
	  extra = node.firstChild.nodeValue;
      }
    if (! lang || ! MIM.imlist[lang])
      return null;
    if (! extra)
      {
	if (! name || ! (im = MIM.imlist[lang][name]))
	  return null;
      }
    else
      {
	if (! (im = MIM.imextra[lang][extra]))
	  return null;
      }
    if (im.load_status != MIM.LoadStatus.Loaded
	&& (im.load_status != MIM.LoadStatus.NotLoaded || ! im.Load ()))
      return null;
    return im;
  }

  var parsers = { };

  parsers['description'] = function (node)
  {
    this.description = node.firstChild.nodeValue;
  }
  parsers['variable-list'] = function (node)
  {
    for (node = node.firstElement (); node; node = node.nextElement ())
      {
	var vname = node.attributes['vname'].nodeValue;
	if (this != MIM.im_global)
	  {
	    var vari = get_global_var (vname);
	    if (vari != null)
	      this.domain.Defvar (vname);
	  }
	vname = Xex.Term.Parse (this.domain, node)
      }
  }
  parsers['command-list'] = function (node)
  {
  }
  parsers['macro-list'] = function (node)
  {
    for (var n = node.firstElement (); n; n = n.nextElement ())
      if (n.nodeName == 'xi:include')
	{
	  var im = include (n);
	  if (! im)
	    alert ('inclusion fail');
	  else
	    for (var macro in im.domain.functions)
	      {
		var func = im.domain.functions[macro];
		if (func instanceof Xex.Macro)
		  im.domain.CopyFunc (this.domain, macro);
	      }
	  n = n.previousSibling;
	  node.removeChild (n.nextSibling);
	}
    Xex.Term.Parse (this.domain, node.firstElement (), null);
  }
  parsers['title'] = function (node)
  {
    this.title = node.firstChild.nodeValue;
  }
  parsers['map-list'] = function (node)
  {
    for (node = node.firstElement (); node; node = node.nextElement ())
      {
	if (node.nodeName == 'xi:include')
	  {
	    var im = include (node);
	    if (! im)
	      {
		alert ('inclusion fail');
		continue;
	      }
	    for (var mapname in im.map_list)
	      this.map_list[mapname] = im.map_list[mapname];
	  }
	else
	  {
	    var map = Xex.Term.Parse (this.domain, node);
	    this.map_list[map.name] = map;
	  }
      }
  }
  parsers['state-list'] = function (node)
  {
    this.domain.map_list = this.map_list;
    for (node = node.firstElement (); node; node = node.nextElement ())
      {
	if (node.nodeName == 'state')
	  {
	    var state = Xex.Term.Parse (this.domain, node);
	    if (! state.title)
	      state.title = this.title;
	    if (! this.initial_state)
	      this.initial_state = state;
	    this.state_list[state.name] = state;
	  }
      }
    delete this.domain.map_list;
  }

  MIM.IM = function (lang, name, extra_id, file)
  {
    this.lang = lang;
    this.name = name;
    this.extra_id = extra_id;
    this.file = file;
    this.load_status = MIM.LoadStatus.NotLoaded;
    this.domain = new Xex.Domain (this.lang + '-'
				  + (this.name != 'nil'
				     ? this.name : this.extra_id),
				  MIM.im_domain, null);
  }

  var proto = {
    Load: function ()
    {
      var node = Xex.Load (null, this.file);
      if (! node)
	{
	  this.load_status = MIM.LoadStatus.Error;
	  return false;
	}
      this.map_list = {};
      this.initial_state = null;
      this.state_list = {};
      for (node = node.firstElement (); node; node = node.nextElement ())
        {
	  var name = node.nodeName;
	  var parser = parsers[name];
	  if (parser)
	    parser.call (this, node);
	}
      this.load_status = MIM.LoadStatus.Loaded;
      return true;
    }
  }

  MIM.IM.prototype = proto;

  MIM.IC = function (im, target)
  {
    if (im.load_status == MIM.LoadStatus.NotLoaded)
      im.Load ();
    if (im.load_status != MIM.LoadStatus.Loaded)
      alert ('im:' + im.name + ' error:' + im.load_status);
    this.im = im;
    this.target = target;
    this.domain = new Xex.Domain ('context', im.domain, this);
    this.active = true;
    this.reset ();
    this.spot = 0;
  }

  MIM.CandidateTable = function ()
  {
    this.table = new Array ();
  }

  MIM.CandidateTable.prototype.get = function (from)
  {
    for (var i = 0; i < this.table.length; i++)
      {
	var elt = this.table[i];
	if (elt.from <= from && elt.to > from)
	  return elt.val;
      }
  }

  MIM.CandidateTable.prototype.put = function (from, to, candidates)
  {
    for (var i = 0; i < this.table.length; i++)
      {
	var elt = this.table[i];
	if (elt.from >= from && elt.from < to
	    || elt.to >= from && elt.to < to)
	  {
	    elt.from = from;
	    elt.to = to;
	    elt.val = candidates;
	    return;
	  }
      }
    this.table.push ({ from: from, to: to, val: candidates });
  }

  MIM.CandidateTable.prototype.adjust = function (from, to, inserted)
  {
    var diff = inserted - (to - from);
    for (var i = 0; i < this.table.length; i++)
      {
	var elt = this.table[i];
	if (elt.from >= to)
	  {
	    elt.from += diff;
	    elt.to += diff;
	  }
      }
  }

  MIM.CandidateTable.prototype.clear = function ()
  {
    this.table.length = 0;
  }

  function Block (index, term)
  {
    this.Index = index;
    if (term.IsStr)
      this.Data = term.val;
    else if (term.IsList)
      {
	this.Data = new Array ();
	for (var i = 0; i < term.val.length; i++)
	  this.Data.push (term.val[i].val);
      }
  }

  Block.prototype.Count = function () { return this.Data.length; }
  Block.prototype.get = function (i)
  {
    return (this.Data instanceof Array ? this.Data[i] : this.Data.charAt (i));
  }

  function fill_group (start)
  {
    var nitems = this.group.length;
    var r = this.row;
    var b = this.blocks[r];

    if (start < b.Index)
      while (start < b.Index)
	b = this.blocks[--r];
    else
      while (start >= b.Index + b.Count ())
	b = this.blocks[++r];
    this.row = r;

    var count = b.Count ();
    start -= b.Index;
    for (var i = 0; i < nitems; i++, start++)
      {
	if (start >= count)
	  {
	    r++;
	    if (r == this.blocks.Length)
	      return i;
	    b = this.blocks[r];
	    count = b.Count ();
	    start = 0;
	  }
	this.group[i] = b.get (start);
      }
    return nitems;
  }

  function Candidates (candidates, column)
  {
    this.column = column;
    this.row = 0;
    this.index = 0;
    this.total = 0;
    this.blocks = new Array ();

    for (var i = 0; i < candidates.length; i++)
      {
	var block = new Block (this.total, candidates[i]);
	this.blocks.push (block);
	this.total += block.Count ();
      }
  }

  Candidates.prototype.Column = function ()
  {
    return (this.column > 0 ? this.index % this.column
	    : this.index - this.blocks[this.row].Index);
  }

  Candidates.prototype.GroupLength = function ()
  {
    if (this.column > 0)
      {
	var start = this.index - (this.index % this.column);
	return (start + this.column <= this.total ? this.column
		: this.total - start);
      }
    return this.blocks[this.row].Count;
  }

  Candidates.prototype.Current = function ()
  {
    var b = this.blocks[this.row];
    return b.get (this.index - b.Index);
  }

  Candidates.prototype.PrevGroup = function ()
  {
    var col = this.Column ();
    var nitems;
    if (this.column > 0)
      {
	this.index -= this.column;
	if (this.index >= 0)
	  nitems = this.column;
	else
	  {
	    var lastcol = (this.total - 1) % this.column;
	    this.index = (col < lastcol ? this.total - lastcol + col
			  : this.total - 1);
	    this.row = this.blocks.length - 1;
	    nitems = lastcol + 1;
	  }
	while (this.blocks[this.row].Index > this.index)
	  this.row--;
      }
    else
      {
	this.row = this.row > 0 ? this.row - 1 : this.blocks.length - 1;
	nitems = this.blocks[this.row].Count ();
	this.index = (this.blocks[this.row].Index
		      + (col < nitems ? col : nitems - 1));
      }
    return nitems;
  }

  Candidates.prototype.NextGroup = function ()
  {
    var col = this.Column ();
    var nitems;
    if (this.column > 0)
      {
	this.index += this.column - col;
	if (this.index < this.total)
	  {
	    if (this.index + col >= this.total)
	      {
		nitems = this.total - this.index;
		this.index = this.total - 1;
	      }
	    else
	      {
		nitems = this.column;
		this.index += col;
	      }
	  }
	else
	  {
	    this.index = col;
	    this.row = 0;
	  }
	while (this.blocks[this.row].Index > this.index)
	  this.row++;
      }
    else
      {
	this.row = this.row < this.blocks.length - 1 ? this.row + 1 : 0;
	nitems = this.blocks[this.row].Count ();
	this.index = (this.blocks[this.row].Index
		      + (col < nitems ? col : nitems - 1));
      }
    return nitems;
  }

  Candidates.prototype.Prev = function ()
  {
    if (this.index == 0)
      {
	this.index = this.total - 1;
	this.row = this.blocks.length - 1;
      }
    else
      {
	this.index--;
	if (this.blocks[this.row].Index > this.index)
	  this.row--;
      }
    }

  Candidates.prototype.Next = function ()
  {
    this.index++;
    if (this.index == this.total)
      {
	this.index = 0;
	this.row = 0;
      }
    else
      {
	var b = this.blocks[this.row];
	if (this.index == b.Index + b.Count ())
	  this.row++;
      }
  }

  Candidates.prototype.First = function ()
  {
    this.index -= this.Column ();
    while (this.blocks[this.row].Index > this.index)
      this.row--;
  }

  Candidates.prototype.Last = function ()
  {
    var b = this.blocks[this.row];
    if (this.column > 0)
      {
	if (this.index + 1 < this.total)
	  {
	    this.index += this.column - this.Column () + 1;
	    while (b.Index + b.Count () <= this.index)
	      b = this.blocks[++this.row];
	  }
      }
    else
      this.index = b.Index + b.Count () - 1;
  }

  Candidates.prototype.Select = function (selector)
  {
    if (selector.type == 'selector')
      {
	switch (selector.val)
	  {
	  case '@<': this.First (); break;
	  case '@>': this.Last (); break;
	  case '@-': this.Prev (); break;
	  case '@+': this.Next (); break;
	  case '@[': this.PrevGroup (); break;
	  case '@]': this.NextGroup (); break;
	  default: break;
	  }
	return this.Current ();
      }
    var col, start, end
    if (this.column > 0)
      {
	col = this.index % this.column;
	start = this.index - col;
	end = start + this.column;
      }
    else
      {
	start = this.blocks[this.row].Index;
	col = this.index - start;
	end = start + this.blocks[this.row].Count;
      }
    if (end > this.total)
      end = this.total;
    this.index += selector.val - col;
    if (this.index >= end)
      this.index = end - 1;
    if (this.column > 0)
      {
	if (selector.val > col)
	  while (this.blocks[this.row].Index + this.blocks[this.row].Count
		 < this.index)
	    this.row++;
	else
	  while (this.blocks[this.row].Index > this.index)
	    this.row--;
      }
    return this.Current ();
  }

  function detach_candidates (ic)
  {
    ic.candidate_table.clear ();
    ic.candidates = null;
    ic.changed |= (MIM.ChangedStatus.Preedit | MIM.ChangedStatus.CursorPos
		   | ChangedStatus.CandidateList
		   | ChangedStatus.CandidateIndex
		   | ChangedStatus.CandidateShow);
  }

  function set_cursor (prefix, pos)
  {
    this.cursor_pos = pos;
    if (pos > 0)
      this.candidates = this.candidate_table.get (pos - 1);
    else
      this.candidates = null;
  }

  function save_state ()
  {
    this.state_var_values = this.domain.SaveValues ();
    this.state_preedit = this.preedit;
    this.state_key_head = this.key_head;
    this.state_pos = this.cursor_pos;
  }

  function restore_state ()
  {
    this.domain.RestoreValues (this.state_var_values);
    this.preedit = this.state_preedit;
    set_cursor.call (this, "restore", this.state_pos);
  }

  function handle_key ()
  {
    var out = this.keymap.Lookup (this.keys, this.key_head);
    var sub = out.map;
    var branch_actions = this.state.keymap.actions;

    MIM.log ('handling ' + this.keys.val[this.key_head]
	     + ' in ' + this.state.name + ':' + this.keymap.name);
    this.key_head = out.index;
    if (sub != this.keymap)
      {

	restore_state.call (this);
	this.keymap = sub;
	MIM.log ('submap found');
	if (this.keymap.actions)
	  {
	    MIM.log ('taking map actions:');
	    if (! this.take_actions (this.keymap.actions))
	      return false;
	  }
	else if (this.keymap.submaps)
	  {
	    MIM.log ('no map actions');
	    for (var i = this.state_key_head; i < this.key_head; i++)
	      {
		MIM.log ('inserting key:' + this.keys.val[i].key);
		this.insert (this.keys.val[i].key, null);
	      }
	  }
	if (! this.keymap.submaps)
	  {
	    MIM.log ('terminal:');
	    if (this.keymap.branch_actions != null)
	      {
		MIM.log ('branch actions:');
		if (! this.take_actions (branch_actions))
		  return false;
	      }
	    if (this.keymap != this.state.keymap)
	      this.shift (this.state);
	  }
      }
    else
      {
	MIM.log ('no submap');
	var current_state = this.state;
	var map = this.keymap;

	if (branch_actions)
	  {
	    MIM.log ('branch actions');
	    if (! this.take_actions (this.keymap.branch_actions))
	      return false;
	  }

	if (map == this.keymap)
	  {
	    MIM.log ('no state change');
	    if (map == this.initial_state.keymap
		&& this.key_head < this.keys.val.length)
	      {
		MIM.log ('unhandled');
		return false;
	      }
	    if (this.keymap != current_state.keymap)
	      this.shift (current_state);
	    else if (this.keymap.actions == null)
	      this.shift (this.initial_state);
	  }
      }
    return true;
  }

  proto = {
    reset: function ()
    {
      this.produced = null;
      this.preedit = '';
      this.preedit_saved = '';
      this.cursor_pos = 0;
      this.marker_positions = {};
      this.candidates = null;
      this.candidate_show = false;
      this.state = null;
      this.prev_state = null;
      this.initial_state = this.im.initial_state;
      this.title = this.initial_state.title;
      this.state_preedit = '';
      this.state_key_head = 0;
      this.state_var_values = {};
      this.state_pos = 0;
      this.keymap = null;
      this.keys = new MIM.KeySeq ();
      this.key_head = 0;
      this.key_unhandled = false;
      this.unhandled_key = null;
      this.changed = MIM.ChangedStatus.None;
      this.error_message = '';
      this.title = this.initial_state.title;
      this.produced = '';
      this.preedit = '';
      this.preedit_saved = '';
      this.marker_positions = {};
      this.candidate_table = new MIM.CandidateTable ();
      this.candidates = null;
      this.candidate_show = false;
      this.shift (this.initial_state);
    },

    catch_args: new Array (Xex.CatchTag._mimtag, null),

    take_actions: function (actions)
    {
      var func_progn = this.domain.GetFunc ('progn');
      var func_catch = this.domain.GetFunc ('catch');
      this.catch_args[1] = new Xex.Funcall (func_progn, null, actions);
      var term = new Xex.Funcall (func_catch, null, this.catch_args);
      term = term.Eval (this.domain);
      return (! term.IsSymbol || term.val != '@mimtag');
    },

    GetSurroundingChar: function (pos)
    {
      if (pos < 0 ? this.caret_pos < - pos : this.target.value.length < pos)
	return 0;
      return this.target.value.charCodeAt (this.caret_pos + pos);
    },
    
    adjust_markers: function (from, to, inserted)
    {
      var diff = inserted - (to - from);

      for (var m in this.marker_positions)
	if (this.marker_positions[m] > from)
	  this.marker_positions[m] = (this.marker_positions[m] >= to
				      ? pos + diff : from);
      if (this.cursor_pos >= to)
	set_cursor.call (this, 'adjust', this.cursor_pos + diff);
      else if (this.cursor_pos > from)
	set_cursor.call (this, 'adjust', from)
    },

    preedit_replace: function (from, to, text, candidates)
    {
      this.preedit = (this.preedit.substring (0, from)
		      + text + this.preedit.substring (to));
      this.adjust_markers (from, to, text.length);
      this.candidate_table.adjust (from, to, text.length);
      if (candidates)
	this.candidate_table.put (from, from + text.length, candidates)
    },

    insert: function (text, candidates)
    {
      this.preedit_replace (this.cursor_pos, this.cursor_pos, text, candidates);
      this.changed = MIM.ChangedStatus.Preedit | MIM.ChangedStatus.CursorPos;
    },

    del: function (pos)
    {
      var deleted = pos - this.cursor_pos;
      if (pos < 0)
	{
	  this.DelSurroundText (pos);
	  pos = 0;
	}
      else if (pos > this.preedit.length)
	{
	  this.DelSurroundText (pos - this.preedit.length);
	  pos = this.preedit.length;
	}
      if  (pos < this.cursor_pos)
	this.preedit = (this.predit.substring (0, pos)
			+ this.preedit.substring (this.cursor_pos));
      else
	this.preedit = (this.preedit.substring (0, this.cursor_pos)
			+ this.predit.substring (pos));
      return deleted;
    },

    show: function ()
    {
      this.candidate_show = true;
      this.changed |= MIM.ChangedStatus.CandidateShow;
    },

    hide: function ()
    {
      this.candidate_show = false;
      this.changed |= MIM.ChangedStatus.CandidateShow;
    },

    move: function (pos)
    {
      if (pos < 0)
	pos = 0;
      else if (pos > this.preedit.length)
	pos = this.preedit.length;
      if (pos != this.cursor_pos)
	{
	  set_cursor.call (this, 'move', pos);
	  this.changed |= MIM.ChangedStatus.Preedit;
	}
    },

    pushback: function (n)
    {
      if (n instanceof MIM.KeySeq)
	{
	  if (this.key_head > 0)
	    this.key_head--;
	  if (this.key_head < this.keys.val.length)
	    this.keys.val.splice (this.key_head,
				  this.keys.val.length - this.key_head);
	  for (var i = 0; i < n.val.length; i++)
	    this.keys.val.push (n.val[i]);
	  return;
	}
      if (n > 0)
	{
	  this.key_head -= n;
	  if (this.key_head < 0)
	    this.key_head = 0;
	}
      else if (n == 0)
	this.key_head = 0;
      else
      {
	this.key_head = - n;
	if (this.key_head > this.keys.val.length)
	  this.key_head = this.keys.val.length;
      }
    },

    pop: function ()
    {
      if (this.key_head < this.keys.val.length)
	this.keys.val.splice (this.key_head, 1);
    },

    commit: function ()
    {
      if (this.preedit.length > 0)
      {
	this.candidate_table.clear ();
	this.produced += this.preedit;
	this.preedit_replace.call (this, 0, this.preedit.length, '', null);
      }
    },

    shift: function (state)
    {
      if (state == null)
        {
	  MIM.log ("shifting back to previous");
	  if (this.prev_state == null)
	    return;
	  state = this.prev_state;
	}
      else
	MIM.log ("shifting to " + state.name);

      if (state == this.initial_state)
        {
	  if (this.state)
	    {
	      this.commit ();
	      this.keys.val.splice (0, this.key_head);
	      this.key_head = 0;
	      this.prev_state = null;
	    }
	}
      else
	{
	  if (state != this.state)
	    this.prev_state = this.state;
	}
      if (state != this.state && state.enter_actions)
	this.take_actions (state.enter_actions);
      if (! this.state || this.state.title != state.title)
	this.changed |= MIM.ChangedStatus.StateTitle;
      this.state = state;
      this.keymap = state.keymap;
      this.state_key_head = this.key_head;
      save_state.call (this);
    },

    Filter: function (key)
    {
      if (! this.active)
	{
	  this.key_unhandled = true;
	  this.unhandled_key = key;
	  return false;
	}
      if (key.key == '_reload')
	return true;
      this.changed = MIM.ChangedStatus.None;
      this.produced = '';
      this.key_unhandled = false;
      this.keys.val.push (key);
      var count = 0;
      while (this.key_head < this.keys.val.length)
	{
	  if (! handle_key.call (this))
	    {
	      if (this.key_head < this.keys.val.length)
		{
		  this.unhandled_key = this.keys.val[this.key_head];
		  this.keys.val.splice (this.key_head, this.key_head + 1);
		}
	      this.key_unhandled = true;
	      break;
	    }
	  if (++count == 10)
	    {
	      this.reset ();
	      this.key_unhandled = true;
	      break;
	    }
	}
      if (this.key_unhandled)
	{
	  this.keys.val.length = 0;
	  this.key_head = this.state_key_head = this.commit_key_head = 0;
	}
      return (! this.key_unhandled
	      && this.produced.length == 0
	      && this.preedit.length == 0);
    }
  }

  MIM.IC.prototype = proto;

  var node = Xex.Load (null, "imlist.xml");
  for (node = node.firstChild; node; node = node.nextSibling)
    if (node.nodeName == 'input-method')
      {
	var lang = null, name = null, extra_id = null, file = null;

	for (var n = node.firstChild; n; n = n.nextSibling)
	  {
	    if (n.nodeName == 'language')
	      lang = n.firstChild.nodeValue;
	    else if (n.nodeName == 'name')
	      name = n.firstChild.nodeValue;
	    else if (n.nodeName == 'extra-id')
	      extra_id = n.firstChild.nodeValue;
	    else if (n.nodeName == 'filename')
	      file = n.firstChild.nodeValue;
	  }
	if (name && name != 'nil')
	  {
	    if (! MIM.imlist[lang])
	      MIM.imlist[lang] = {};
	    MIM.imlist[lang][name] = new MIM.IM (lang, name, extra_id, file);
	  }
	else if (extra_id && extra_id != 'nil')
	  {
	    if (! MIM.imextra[lang])
	      MIM.imextra[lang] = {};
	    MIM.imextra[lang][extra_id] = new MIM.IM (lang, name, extra_id, file);
	  }
      }
  if (MIM.imextra.t && MIM.imextra.t.global)
    MIM.im_global = MIM.imextra.t.global;
  else
    {
      MIM.im_global = new MIM.IM ('t', 'nil', 'global', null);
      MIM.im_global.load_status = MIM.LoadStatus.Error;
    }
  node = undefined;
}) ();

(function () {
  var keys = new Array ();
  keys[0x09] = 'tab';
  keys[0x08] = 'backspace';
  keys[0x0D] = 'return';
  keys[0x1B] = 'escape';
  keys[0x20] = 'space';
  keys[0x21] = 'pageup';
  keys[0x22] = 'pagedown';
  keys[0x23] = 'end';
  keys[0x24] = 'home';
  keys[0x25] = 'left';
  keys[0x26] = 'up';
  keys[0x27] = 'right';
  keys[0x28] = 'down';
  keys[0x2D] = 'insert';
  keys[0x2E] = 'delete';
  for (var i = 1; i <= 12; i++)
    keys[111 + i] = "f" + i;
  keys[0x90] = "numlock";
  keys[0xF0] = "capslock";

  MIM.decode_key_event = function (event)
  {
    var key = ((event.type == 'keydown' || event.keyCode) ? event.keyCode
	       : event.charCode ? event.charCode
	       : false);
    if (! key)
      return false;
    if (event.type == 'keydown')
      {
	key = keys[key];
	if (! key)
	  return false;
	if (event.shiftKey) key = "S-" + key ;
      }
    else
      key = String.fromCharCode (key);
    if (event.altKey) key = "A-" + key ;
    if (event.ctrlKey) key = "C-" + key ;
    return new MIM.Key (key);
  }
}) ();

MIM.add_event_listener
  = (window.addEventListener
     ? function (target, type, listener) {
       target.addEventListener (type, listener, false);
     }
     : window.attachEvent
     ? function (target, type, listener) {
       target.attachEvent ('on' + type,
			   function() {
			     listener.call (target, window.event);
			   });
     }
     : function (target, type, listener) {
       target['on' + type]
	 = function (e) { listener.call (target, e || window.event); };
     });

MIM.log = function (msg)
{
  var node = document.getElementById ('log');
  node.value = msg + "\n" + node.value;
}

MIM.debug_print = function (event, ic)
{
  if (! MIM.debug)
    return;
  if (! MIM.debug_nodes)
    {
      MIM.debug_nodes = new Array ();
      MIM.debug_nodes['keydown'] = document.getElementById ('keydown');
      MIM.debug_nodes['keypress'] = document.getElementById ('keypress');
      MIM.debug_nodes['status0'] = document.getElementById ('status0');
      MIM.debug_nodes['status1'] = document.getElementById ('status1');
      MIM.debug_nodes['keymap0'] = document.getElementById ('keymap0');
      MIM.debug_nodes['keymap1'] = document.getElementById ('keymap1');
      MIM.debug_nodes['preedit0'] = document.getElementById ('preedit0');
      MIM.debug_nodes['preedit1'] = document.getElementById ('preedit1');
    }
  var target = event.target;
  var code = event.keyCode;
  var ch = event.type == 'keydown' ? 0 : event.charCode;
  var key = MIM.decode_key_event (event);
  var index;

  MIM.debug_nodes[event.type].innerHTML = "" + code + "/" + ch + " : " + key;
  index = (event.type == 'keydown' ? '0' : '1');
  if (ic)
    MIM.debug_nodes['status' + index].innerHTML = ic.im.load_status;
  else
    MIM.debug_nodes['status' + index].innerHTML = 'no IM';
  MIM.debug_nodes['keymap' + index].innerHTML = ic.state.name;
  MIM.debug_nodes['preedit' + index].innerHTML = ic.preedit;
};

MIM.get_range = function (target, range)
{
  if (target.selectionStart != null) // for Mozilla
    {
      range[0] = target.selectionStart;
      range[1] = target.selectionEnd;
    }
  else				// for IE
    {
      var r = document.selection.createRange ();
      var rr = r.duplicate ();

      rr.moveToElementText (target);
      rr.setEndPoint ('EndToEnd', range);
      range[0] = rr.text.length - r.text.length;
      range[1] = rr.text.length;
    }
}

MIM.set_caret = function (target, ic)
{
  if (target.setSelectionRange)	// Mozilla
    {
      var scrollTop = target.scrollTop;
      target.setSelectionRange (ic.spot, ic.spot + ic.preedit.length);
      target.scrollTop = scrollTop;
    }
  else				// IE
    {
      var range = target.createTextRange ();
      range.moveStart ('character', ic.spot);
      range.moveEnd ('character', ic.spot + ic.preedit.length);
      range.select ();
    }
};

(function () {
  var range = new Array ();

  MIM.check_range = function (target, ic)
  {
    MIM.get_range (target, range);
    if (range[0] != ic.spot || range[1] - range[0] != ic.preedit.length
	|| target.value.substring (range[0], range[1]) != ic.preedit)
      {
	MIM.log ('reset:' + ic.spot + '-' + (ic.spot + ic.preedit.length)
		 + '/' + range[0] + '-' + range[1]);
	ic.reset ();
      }
    target.value = (target.value.substring (0, range[0])
		    + target.value.substring (range[1]));
    ic.spot = range[0];
  }
}) ();

MIM.update = function (target, ic)
{
  var text = target.value;
  target.value = (text.substring (0, ic.spot)
		  + ic.produced
		  + ic.preedit
		  + text.substring (ic.spot));
  ic.spot += ic.produced.length;
  MIM.set_caret (target, ic);
};

MIM.reset_ic = function (event)
{
  if (event.target.mim_ic)
    {
      var ic = event.target.mim_ic;
      var pos = ic.spot + ic.preedit.length;
      ic.reset ();
      if (pos > ic.spot)
	event.target.setSelectionRange (pos, pos);
    }
};

MIM.keydown = function (event)
{
  var target = event.target;
  if (! (target.type == "text" || target.type == "textarea"))
    return;

  var ic = target.mim_ic;
  if (! ic || ic.im != MIM.current)
    {
      MIM.log ('creating IC');
      ic = new MIM.IC (MIM.current, target);
      target.mim_ic = ic;
      MIM.add_event_listener (target, 'blur', MIM.reset_ic);
    }
  if (ic.im.load_status != MIM.LoadStatus.Loaded)
    return;
  MIM.check_range (target, ic);
  MIM.debug_print (event, ic);
  ic.key = MIM.decode_key_event (event);
};

MIM.keypress = function (event)
{
  if (! (event.target.type == "text" || event.target.type == "textarea"))
    return;

  var ic = event.target.mim_ic;
  var i;

  try {
    if (ic.im.load_status != MIM.LoadStatus.Loaded)
      return;
    if (! ic.key)
      ic.key = MIM.decode_key_event (event);
    if (! ic.key)
      {
	ic.reset ();
	return;
      }
    
    MIM.log ("filtering " + ic.key);
    var result = ic.Filter (ic.key);
    MIM.update (event.target, ic);
    if (! ic.key_unhandled)
      event.preventDefault ();
  } finally {
    MIM.debug_print (event, ic);
  }
  return;
};

MIM.select_im = function (event)
{
  var target = event.target.parentNode;
  while (target.tagName != "SELECT")
    target = target.parentNode;
  var idx = 0;
  var im = false;
  for (var lang in MIM.imlist)
    for (var name in MIM.imlist[lang])
      if (idx++ == target.selectedIndex)
	{
	  im = MIM.imlist[lang][name];
	  break;
	}
  document.getElementsByTagName ('body')[0].removeChild (target);
  target.target.focus ();
  if (im && im != MIM.current)
    {
      MIM.current = im;
      MIM.log ('select IM: ' + im.name);
    }
};

MIM.destroy_menu = function (event)
{
  if (event.target.tagName == "SELECT")
    document.getElementsByTagName ('body')[0].removeChild (event.target);
};

MIM.select_menu = function (event)
{
  var target = event.target;

  if (! ((target.type == "text" || target.type == "textarea")
	 && event.which == 1 && event.ctrlKey))
    return;

  var sel = document.createElement ('select');
  sel.onclick = MIM.select_im;
  sel.onmouseout = MIM.destroy_menu;
  sel.style.position='absolute';
  sel.style.left = (event.clientX - 10) + "px";
  sel.style.top = (event.clientY - 10) + "px";
  sel.target = target;
  var idx = 0;
  for (var lang in MIM.imlist)
    for (var name in MIM.imlist[lang])
      {
	var option = document.createElement ('option');
	var imname = lang + "-" + name;
	option.appendChild (document.createTextNode (imname));
	option.value = imname;
	sel.appendChild (option);
	if (MIM.imlist[lang][name] == MIM.current)
	  sel.selectedIndex = idx;
	idx++;
      }
  sel.size = idx;
  document.getElementsByTagName ('body')[0].appendChild (sel);
};

MIM.test = function ()
{
  var im = MIM.imlist['t']['latn-post'];
  var ic = new MIM.IC (im, null);

  ic.Filter (new MIM.Key ('a'));
  ic.Filter (new MIM.Key ("'"));

  if (true)
    document.getElementById ('text').value = ic.produced + ic.preedit;
  else {
    try {
      document.getElementById ('text').value
	= Xex.Term.Parse (domain, body).Eval (domain).toString ();
    } catch (e) {
      if (e instanceof Xex.ErrTerm)
	alert (e);
      throw e;
    }
  }
}


MIM.init = function ()
{
  MIM.add_event_listener (window, 'keydown', MIM.keydown);
  MIM.add_event_listener (window, 'keypress', MIM.keypress);
  MIM.add_event_listener (window, 'mousedown', MIM.select_menu);
  if (window.location == 'http://localhost/mim/index.html')
    MIM.server = 'http://localhost/mim';
  MIM.current = MIM.imlist['vi']['telex'];
};

MIM.init_debug = function ()
{
  MIM.debug = true;
  MIM.init ();
};
